import unittest
from mindsdb.integrations.handlers.huggingface_api_handler.huggingface_api_handler import HuggingFaceInferenceAPIHandler
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE


class HuggingFaceAPIHandlerTest(unittest.TestCase):
    pass


if __name__ == '__main__':
    unittest.main()
